/*
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2001 The Ant-Contrib project.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "The Jakarta Project", "Ant", and "Apache Software
 *    Foundation" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */
package net.sf.antcontrib.cpptasks;

import java.util.Enumeration;
import java.util.Vector;

import net.sf.antcontrib.cpptasks.compiler.CommandLineCompiler;
import net.sf.antcontrib.cpptasks.compiler.Compiler;
import net.sf.antcontrib.cpptasks.compiler.Processor;
import net.sf.antcontrib.cpptasks.gcc.GccCCompiler;
import net.sf.antcontrib.cpptasks.types.CompilerArgument;
import net.sf.antcontrib.cpptasks.types.ConditionalPath;
import net.sf.antcontrib.cpptasks.types.IncludePath;
import net.sf.antcontrib.cpptasks.types.SystemIncludePath;
import net.sf.antcontrib.cpptasks.types.UndefineArgument;
import net.sf.antcontrib.cpptasks.types.DefineSet;
import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.types.EnumeratedAttribute;


/**
 * A compiler definition.  compiler elements may be placed
 *    either as children of a cc element or the project element.
 *    A compiler element with an id attribute may be referenced
 *    from compiler elements with refid or extends attributes.
 *
 * @author Adam Murdoch
 */
public final class CompilerDef extends ProcessorDef
{

    /** The source file sets. */
    private final Vector defineSets = new Vector();
    private final Vector includePaths = new Vector();
    private final Vector sysIncludePaths = new Vector();
    private Boolean multithreaded;
    private Boolean exceptions;
    private int warnings = -1;
    private final Vector precompileDefs = new Vector();


    public CompilerDef() {
    }

    public Processor getProcessor() {
      Processor processor = super.getProcessor();
      if(processor == null) {
        processor = GccCCompiler.getInstance();
      }
      if (getLibtool() && processor instanceof CommandLineCompiler) {
        CommandLineCompiler compiler = (CommandLineCompiler) processor;
        processor = compiler.getLibtoolCompiler();
      }
      return processor;
    }

    /**
     * Specifies precompilation prototype file and exclusions. 
     *
     */
    public PrecompileDef createPrecompile() throws BuildException {
        if ( isReference() ) {
            throw noChildrenAllowed();
        }
        PrecompileDef precomp = new PrecompileDef();
        precomp.setProject(project);
        precompileDefs.addElement(precomp);
        return precomp;
    }

    public PrecompileDef getActivePrecompile(CompilerDef ccElement) {
      if( isReference()) {
          return ((CompilerDef)getCheckedRef(CompilerDef.class, "CompilerDef")).
                getActivePrecompile(ccElement);
      }
      PrecompileDef current = null;
      Enumeration enum = precompileDefs.elements();
      while(enum.hasMoreElements()) {
        current = (PrecompileDef) enum.nextElement();
        if (current.isActive()) {
          return current;
        }
      }
      CompilerDef extendedDef = (CompilerDef) getExtends();
      if (extendedDef != null) {
        current = extendedDef.getActivePrecompile(null);
        if (current != null) {
          return current;
        }
      }
      if(ccElement != null && getInherit()) {
        return ccElement.getActivePrecompile(null);
      }
      return null;
    }




    /**
     * Adds a compiler command-line arg.
     */
    public void addConfiguredCompilerArg(CompilerArgument arg)
    {
        if ( isReference() ) {
            throw noChildrenAllowed();
        }
        addConfiguredProcessorArg(arg);
    }

    /**
     * Adds a defineset.
     */
    public void addConfiguredDefineset(DefineSet defs)
    {
        if(defs == null) {
          throw new NullPointerException("defs");
        }
        if ( isReference() ) {
            throw noChildrenAllowed();
        }
        defineSets.addElement(defs);
    }



     /**
     * Creates an include path.
     */
    public IncludePath createIncludePath()
    {
        if(project == null) {
          throw new java.lang.IllegalStateException("project must be set");
        }
        if ( isReference() ) {
            throw noChildrenAllowed();
        }
        IncludePath path = new IncludePath(project);
        includePaths.addElement(path);
        return path;
    }

    /**
     * Creates a system include path.  Locations and timestamps of
     *    files located using the system include paths are not
     *    used in dependency analysis.
     *
     *
     *    Standard include locations should not be specified.
     *    The compiler adapters should recognized the
     *    settings from the appropriate environment variables
     *    or configuration files.
     */
    public SystemIncludePath createSysIncludePath()
    {
        if(project == null) {
          throw new java.lang.IllegalStateException("project must be set");
        }
        if ( isReference() ) {
            throw noChildrenAllowed();
        }
        SystemIncludePath path = new SystemIncludePath(project);
        sysIncludePaths.addElement(path);
        return path;
    }



    /**
     * Returns the compiler-specific include path.
     */
    public String[] getActiveIncludePaths() {
      if(isReference()) {
          return ((CompilerDef)getCheckedRef(CompilerDef.class, "CompilerDef")).
                getActiveIncludePaths();
      }
      return getActivePaths(includePaths);
    }

    public String[] getActiveSysIncludePaths() {
      if(isReference()) {
          return ((CompilerDef)getCheckedRef(CompilerDef.class, "CompilerDef")).
                getActiveSysIncludePaths();
      }
        return getActivePaths(sysIncludePaths);
    }

    /**
     * Enables or disables generation of multithreaded code.
     * Unless specified, multithreaded code generation is enabled.
     *
     * @param multi If true, generated code may be multithreaded.
     */
    public void setMultithreaded(boolean multithreaded) {
      if ( isReference() ) {
            throw tooManyAttributes();
      }
      this.multithreaded = booleanValueOf(multithreaded);
    }


    public boolean getMultithreaded(CompilerDef[] defaultProviders, int index) {
      if(isReference()) {
        return ((CompilerDef)getCheckedRef(CompilerDef.class, "CompilerDef")).
                getMultithreaded(defaultProviders,index);
      }
      if(multithreaded != null) {
        return multithreaded.booleanValue();
      }
      else {
        if(defaultProviders != null && index < defaultProviders.length) {
          return defaultProviders[index].getMultithreaded(defaultProviders,index+1);
        }
      }
      return true;
    }

    /**
     * Enables or disables exception support.
     *
     * @param exceptions if true, exceptions are supported.
     *
     */
    public void setExceptions(boolean exceptions) {
      if ( isReference() ) {
            throw tooManyAttributes();
      }
      this.exceptions = booleanValueOf(exceptions);
    }


    public final boolean getExceptions(CompilerDef[] defaultProviders, int index) {
      if(isReference()) {
        return ((CompilerDef)getCheckedRef(CompilerDef.class, "CompilerDef")).
                getExceptions(defaultProviders,index);
      }
      if(exceptions != null) {
        return exceptions.booleanValue();
      }
      else {
        if(defaultProviders != null && index < defaultProviders.length) {
          return defaultProviders[index].getExceptions(defaultProviders,index+1);
        }
      }
      return false;
    }


    private String[] getActivePaths(Vector paths) {
        if(project == null) {
          throw new java.lang.IllegalStateException("project not set");
        }
        Vector activePaths = new Vector(paths.size());
        for(int i = 0; i < paths.size(); i++) {
          ConditionalPath path = (ConditionalPath) paths.elementAt(i);
          if(path.isActive(project)) {
            String[] pathEntries = path.list();
            for(int j = 0; j < pathEntries.length; j++) {
              activePaths.addElement(pathEntries[j]);
            }
          }
        }
        String[] pathNames = new String[activePaths.size()];
        activePaths.copyInto(pathNames);
        return pathNames;
      }



    /**
     * Sets compiler type.
     *
     *
     * <table width="100%" border="1">
     *    <thead>Supported compilers</thead>
     *    <tr><td>gcc (default)</td><td>GCC C++ compiler</td></tr>
     *    <tr><td>g++</td><td>GCC C++ compiler</td></tr>
     *    <tr><td>c++</td><td>GCC C++ compiler</td></tr>
     *   <tr><td>msvc</td><td>Microsoft Visual C++</td></tr>
     *   <tr><td>bcc</td><td>Borland C++ Compiler</td></tr>
     *   <tr><td>msrc</td><td>Microsoft Resource Compiler</td></tr>
     *   <tr><td>brc</td><td>Borland Resource Compiler</td></tr>
     *   <tr><td>df</td><td>Compaq Visual Fortran Compiler</td></tr>
     *   <tr><td>midl</td><td>Microsoft MIDL Compiler</td></tr>
     *   <tr><td>icl</td><td>Intel C++ compiler for Windows (IA-32)</td></tr>
     *   <tr><td>ecl</td><td>Intel C++ compiler for Windows (IA-64)</td></tr>
     *   <tr><td>icc</td><td>Intel C++ compiler for Linux (IA-32)</td></tr>
     *   <tr><td>ecc</td><td>Intel C++ compiler for Linux (IA-64)</td></tr>
     *   <tr><td>CC</td><td>Sun ONE C++ compiler</td></tr>
     *   <tr><td>aCC</td><td>HP aC++ C++ Compiler</td></tr>
     *   <tr><td>os390</td><td>OS390 C Compiler</td></tr>
     *   <tr><td>os400</td><td>Icc Compiler</td></tr>
     *   <tr><td>sunc89</td><td>Sun C89 C Compiler</td></tr>
     *   <tr><td>xlC</td><td>VisualAge C Compiler</td></tr>
     * </table>
     *
     */
    public void setName(CompilerEnum name) throws BuildException {
        if ( isReference() ) {
            throw tooManyAttributes();
        }
        Compiler compiler = name.getCompiler();
        setProcessor(compiler);
    }

    /**
     *   Sets the default compiler adapter.  Use the "name" attribute
     *   when the compiler is a supported compiler.
     *
     *   @param classname fully qualified classname which implements CompilerAdapter
     */
    public void setClassname(String classname) throws BuildException {
        if ( isReference() ) {
            throw tooManyAttributes();
        }
        super.setClassname(classname);
        Processor proc = getProcessor();
        if(!(proc instanceof Compiler)) {
           throw new BuildException(classname + " does not implement Compiler");
        }
    }

    protected void setProcessor(Processor proc) throws BuildException {
        try {
            super.setProcessor((Compiler) proc);
        }
        catch(ClassCastException ex) {
            throw new BuildException(ex);
        }
    }



    public UndefineArgument[] getActiveDefines() {
      if(project == null) {
        throw new java.lang.IllegalStateException("project must be set before this call");
      }
      if(isReference()) {
          return ((CompilerDef)getCheckedRef(CompilerDef.class, "CompilerDef")).
                getActiveDefines();
      }
      Vector actives = new Vector();
      for(int i = 0; i < defineSets.size(); i++) {
        DefineSet currentSet = (DefineSet) defineSets.elementAt(i);
        UndefineArgument[] defines = currentSet.getDefines();
        for(int j = 0; j < defines.length; j++) {
          if(defines[j].isActive(project)) {
            actives.addElement(defines[j]);
          }
        }
      }
      UndefineArgument[] retval = new UndefineArgument[actives.size()];
      actives.copyInto(retval);
      return retval;
    }


    /**
     * Enumerated attribute with the values "none", "severe",
     *    "default", "production", "diagnostic",
     *    and "failtask".
     */
    public static class WarningLevel extends EnumeratedAttribute {
        public String[] getValues() {
            return new String[] {"none", "severe", "default",
                "production", "diagnostic", "aserror" };
        }
    }

    /**
     * Enumerated attribute with the values "none", "severe",
     *    "default", "production", "diagnostic",
     *    and "failtask".
     */
    public void setWarnings(CompilerDef.WarningLevel level) {
      warnings = level.getIndex();
    }

    public int getWarnings(CompilerDef[] defaultProviders, int index) {
      if(isReference()) {
          return ((CompilerDef)getCheckedRef(CompilerDef.class, "CompilerDef")).
                getWarnings(defaultProviders,index);
      }
      if(warnings == -1) {
        if(defaultProviders != null && index < defaultProviders.length) {
          return defaultProviders[index].getWarnings(defaultProviders,index+1);
        }
      }
      return warnings;
    }

    public void execute() throws org.apache.tools.ant.BuildException
    {
        throw new org.apache.tools.ant.BuildException("Not an actual task, but looks like one for documentation purposes");
    }




}
