/*
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2002 The Ant-Contrib project.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "The Jakarta Project", "Ant", and "Apache Software
 *    Foundation" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */
package net.sf.antcontrib.cpptasks;

import java.io.File;
import java.util.Enumeration;
import java.util.Vector;

import net.sf.antcontrib.cpptasks.types.ConditionalFileSet;
import org.apache.tools.ant.DirectoryScanner;
import org.apache.tools.ant.types.DataType;





/**
 * An element that specifies a prototype file and
 * rules for source files that should not use
 * precompiled headers
 *
 * @author Curt Arnold
 */
public final class PrecompileDef extends DataType {

    /**
     *   Directory of prototype file
     */
    private File prototype = new File("stdafx.cpp");

    private final Vector exceptSets = new Vector();

    private String ifCond;

    private String unlessCond;


    /**
     *  Constructor
     *
     */
    public PrecompileDef() {
    }

    public boolean isActive() {
      boolean isActive = CUtil.isActive(project, ifCond, unlessCond);
      if (!isActive) {
        PrecompileDef ref = getRef();
        if (ref != null) {
          return ref.isActive();
        }
      }
      return isActive;
    }


    private PrecompileDef getRef() {
      if( isReference()) {
          return ((PrecompileDef)getCheckedRef(PrecompileDef.class,
                "PrecompileDef"));
      }
      return null;
    }

    /**
     * Sets file to precompile.
     *
     * Should be a source file that includes only one
     * unguarded header file.  Default value is "stdafx.cpp".
     *
     * @param prototype file path for prototype source file
     */
     public void setPrototype(File prototype) {
        if (isReference()) {
          throw tooManyAttributes();
        }
        if (prototype == null) {
          throw new NullPointerException("prototype");
        }
        this.prototype = prototype;
     }

     /**
      * Gets prototype source file
      *
      */
      public File getPrototype() {
        PrecompileDef ref = getRef();
        if (ref != null) {
          return ref.getPrototype();
        }
        return prototype;
      }


    /**
     * Adds filesets that specify files that should not be processed
     * with precompiled headers enabled.
     *
     * @param exceptSet FileSet specify files that should not be
     * processed with precompiled headers enabled.
     */
    public PrecompileExceptDef createExcept() {
        return new PrecompileExceptDef(this);
    }

    /**
     * Method used by PrecompileExceptDef to add
     * exception set to PrecompileDef.
     */
    public void appendExceptFileSet(ConditionalFileSet exceptSet) {
      exceptSet.setProject(getProject());
      exceptSets.addElement(exceptSet);
    }

    public String[] getExceptFiles() {
      PrecompileDef ref = getRef();
      if (ref != null) {
        return ref.getExceptFiles();
      }
      if (exceptSets.size() == 0) {
        return new String[0];
      }
      String[] exceptFiles = null;
      Enumeration setEnum = exceptSets.elements();
      while (setEnum.hasMoreElements()) {
        ConditionalFileSet exceptSet = (ConditionalFileSet) setEnum.nextElement();
        if (exceptSet.isActive()) {
          DirectoryScanner scanner = exceptSet.getDirectoryScanner(project);
          String[] scannerFiles = scanner.getIncludedFiles();
          if (exceptFiles == null) {
            exceptFiles = scannerFiles;
          } else {
            if (scannerFiles.length > 0) {
              String[] newFiles = new String[exceptFiles.length + scannerFiles.length];
              for (int i = 0; i < exceptFiles.length; i++) {
                newFiles[i] = exceptFiles[i];
              }
              int index = exceptFiles.length;
              for (int i = 0; i < scannerFiles.length; i++) {
                newFiles[index++] = scannerFiles[i];
              }
              exceptFiles = newFiles;
            }
          }
        }
      }
      if (exceptFiles == null) {
        exceptFiles = new String[0];
      }
      return exceptFiles;
    }

    /**
     *  Set the 'if' condition.
     *
     *  The processor will be ignored unless the property
     *    is defined.
     *
     *  The value of property is insignificant, but values
     *      that would imply misinterpretation ("false", "no")
     *      will throw an exception when isActive() is evaluated.
     *
     *  @param propName name of property
     */
    public void setIf(String propName) {
        ifCond = propName;
    }

    /**
     *  Set the 'unless' condition. If named property exists at
     *    execution time, the processor will be ignored.
     *
     *    Value of property is insignificant, but values
     *      that would imply misinterpretation ("false", "no")
     *      of the behavior will throw an exception when isActive is called.
     *
     *  @param propName name of property
     */
    public void setUnless(String propName) {
        unlessCond = propName;
    }


    public void execute() throws org.apache.tools.ant.BuildException
    {
        throw new org.apache.tools.ant.BuildException("Not an actual task, but looks like one for documentation purposes");
    }

    /**
     *    Specifies that this element should behave as if
     *      the content of the element with the matching id attribute
     *      was inserted at this location.
     *
     *    @param ref Reference to other element
     *
     */
    public void setRefid(org.apache.tools.ant.types.Reference ref) {
        super.setRefid(ref);
    }


    /**
     *    Sets an id that can be used to reference this element.
     *
     *    @param id id
     */
    public void setId(String id) {
        //
        //  this is actually accomplished by a different
        //     mechanism, but we can document it 
        //
    }
    
    /** 
     * Sets a description of the current data type.
     */
    public void setDescription(String desc) {
        super.setDescription(desc);
    }


}
