/*
 * The Apache Software License, Version 1.1l
 *
 * Copyright (c) 2002 The Ant-Contrib project.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "The Jakarta Project", "Ant", and "Apache Software
 *    Foundation" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */
package net.sf.antcontrib.cpptasks;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.UnsupportedEncodingException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import net.sf.antcontrib.cpptasks.compiler.ProcessorConfiguration;
import org.apache.tools.ant.BuildException;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

/**
 * A history of the compiler and linker settings used to build
 * the files in the same directory as the history.
 *
 * @author Curt Arnold
 */
public final class TargetHistoryTable  {

    /** Flag indicating whether the cache should be written back to file. */
    private boolean dirty;

    /** The file the cache was loaded from. */
    private /* final */ File historyFile;

    /**
     * a hashtable of TargetHistory's keyed by output file name
     */
    private final Hashtable history = new Hashtable();

    private /* final */ File outputDir;
    private String outputDirPath;


      /**
       * Creates a target history table from history.xml in the
       * output directory, if it exists.  Otherwise, initializes
       * the history table empty.
       *
       * @param task task used for logging history load errors
       * @param outputDir output directory for task
       */
    public TargetHistoryTable(CCTask task,File outputDir) throws BuildException {
       if(outputDir == null) {
          throw new NullPointerException("outputDir");
       }
       if(!outputDir.isDirectory()) {
          throw new BuildException("Output directory is not a directory");
       }
       if(!outputDir.exists()) {
          throw new BuildException("Output directory does not exist");
       }
       this.outputDir = outputDir;
       try {
          outputDirPath = outputDir.getCanonicalPath();
       }
       catch(IOException ex) {
          outputDirPath = outputDir.toString();
       }
       //
       //   load any existing history from file
       //       suppressing any records whose corresponding
       //       file does not exist, is zero-length or
       //          last modified dates differ
       historyFile = new File(outputDir,"history.xml");
       if(historyFile.exists()) {
          SAXParserFactory factory = SAXParserFactory.newInstance();
          factory.setValidating(false);
          try {
            SAXParser parser = factory.newSAXParser();
            parser.parse(historyFile, new TargetHistoryTableHandler(history,outputDir));
          }
          catch(Exception ex) {
            //
            //   a failure on loading this history is not critical
            //       but should be logged
            task.log("Error reading history.xml: " + ex.toString());
          }
       }
    }

    public TargetHistory get(String configId, String outputName) {
      TargetHistory targetHistory = (TargetHistory) history.get(outputName);
      if(targetHistory != null) {
        if(!targetHistory.getProcessorConfiguration().equals(configId)) {
          targetHistory = null;
        }
      }
      return targetHistory;
    }

    public void update(long startTime, ProcessorConfiguration config, String[] sources) {
      String configId = config.getIdentifier();
      String[] onesource = new String[1];
      String outputName;
      for(int i = 0; i < sources.length; i++) {
        onesource[0] = sources[i];
        outputName = config.getOutputFileName(sources[i]);
        update(startTime,configId,outputName,onesource);
      }
    }

    private void update(long startTime,String configId, String outputName, String[] sources) {
      File outputFile = new File(outputDir,outputName);
      //
      //   if output file doesn't exist or predates the start of the
      //        compile step (most likely a compilation error) then
      //        do not write add a history entry
      //
      if(outputFile.exists() && outputFile.lastModified() > startTime) {
        dirty = true;
        history.remove(outputName);
        SourceHistory[] sourceHistories = new SourceHistory[sources.length];
        for(int i = 0; i < sources.length; i++) {
           File sourceFile = new File(sources[i]);
           long lastModified = sourceFile.lastModified();
           String relativePath = CUtil.getRelativePath(outputDirPath,sourceFile);
           sourceHistories[i] = new SourceHistory(relativePath,lastModified);
        }
        TargetHistory newHistory = new TargetHistory(configId,
          outputName,outputFile.lastModified(),sourceHistories);
        history.put(outputName,newHistory);
      }
    }

    public void update(long startTime,TargetInfo linkTarget) {
      File outputFile = linkTarget.getOutput();
      String outputName = outputFile.getName();

      //
      //   if output file doesn't exist or predates the start of the
      //        compile or link step (most likely a compilation error) then
      //        do not write add a history entry
      //
      if(outputFile.exists() && outputFile.lastModified() > startTime) {
        dirty = true;
        history.remove(outputName);
        SourceHistory[] sourceHistories = linkTarget.getSourceHistories(outputDirPath);
        TargetHistory newHistory = new TargetHistory(
          linkTarget.getConfiguration().getIdentifier(),
          outputName,outputFile.lastModified(),
          sourceHistories);
        history.put(outputName,newHistory);
      }
    }


    public void commit() throws IOException {
      //
      //   if not dirty, no need to update file
      //
      if(dirty) {
        //
        //   build (small) hashtable of config id's in history
        //
        Hashtable configs = new Hashtable(20);
        Enumeration elements = history.elements();
        while(elements.hasMoreElements()) {
          TargetHistory targetHistory = (TargetHistory) elements.nextElement();
          String configId = targetHistory.getProcessorConfiguration();
          if(configs.get(configId) == null) {
            configs.put(configId,configId);
          }
        }

        FileOutputStream outStream = new FileOutputStream(historyFile);
        OutputStreamWriter outWriter;
        //
        //   early VM's don't support UTF-8 encoding
        //       try and fallback to the default encoding
        //           otherwise
        String encodingName = "UTF-8";
        try {
          outWriter = new OutputStreamWriter(outStream,"UTF-8");
        }
        catch(UnsupportedEncodingException ex) {
          outWriter = new OutputStreamWriter(outStream);
          encodingName = outWriter.getEncoding();
        }
        BufferedWriter writer = new BufferedWriter(outWriter);
        writer.write("<?xml version='1.0' encoding='");
        writer.write(encodingName);
        writer.write("'?>\n");
        writer.write("<history>\n");
        StringBuffer buf = new StringBuffer(200);
        Enumeration configEnum = configs.elements();
        while(configEnum.hasMoreElements()) {
          String configId = (String) configEnum.nextElement();
          buf.setLength(0);
          buf.append("   <processor signature=\"");
          buf.append(CUtil.xmlAttribEncode(configId));
          buf.append("\">\n");
          writer.write(buf.toString());
          elements = history.elements();
          while(elements.hasMoreElements()) {
            TargetHistory targetHistory = (TargetHistory) elements.nextElement();
            if(targetHistory.getProcessorConfiguration().equals(configId)) {
              buf.setLength(0);
              buf.append("      <target file=\"");
              buf.append(CUtil.xmlAttribEncode(targetHistory.getOutput()));
              buf.append("\" lastModified=\"");
              buf.append(Long.toHexString(targetHistory.getOutputLastModified()));
              buf.append("\">\n");
              writer.write(buf.toString());
              SourceHistory[] sourceHistories = targetHistory.getSources();
              for(int i = 0; i < sourceHistories.length; i++) {
                buf.setLength(0);
                buf.append("         <source file=\"");
                buf.append(CUtil.xmlAttribEncode(sourceHistories[i].getRelativePath()));
                buf.append("\" lastModified=\"");
                buf.append(Long.toHexString(sourceHistories[i].getLastModified()));
                buf.append("\"/>\n");
                writer.write(buf.toString());
              }
              writer.write("      </target>\n");
            }
          }
          writer.write("   </processor>\n");
        }
        writer.write("</history>\n");
        writer.close();
        dirty = false;
      }
    }


    public void markForRebuild(Hashtable targetInfos) {
      Enumeration targetInfoEnum = targetInfos.elements();
      while(targetInfoEnum.hasMoreElements()) {
        markForRebuild((TargetInfo) targetInfoEnum.nextElement());
      }
   }


    public void markForRebuild(TargetInfo targetInfo) {
        //
        //     if it must already be rebuilt, no need to check further
        //
        if(!targetInfo.getRebuild())  {
          TargetHistory history = get(
            targetInfo.getConfiguration().toString(),
            targetInfo.getOutput().getName());
          if(history == null) {
            targetInfo.mustRebuild();
          }
          else {
            SourceHistory[] sourceHistories = history.getSources();
            File[] sources = targetInfo.getSources();
            if(sourceHistories.length != sources.length) {
              targetInfo.mustRebuild();
            }
            else {
              for(int i = 0; i < sourceHistories.length && !targetInfo.getRebuild(); i++) {
                //
                //   relative file name, must absolutize it on output directory
                //
                  boolean foundMatch = false;
                    String historySourcePath = sourceHistories[i].getAbsolutePath(outputDir);
                    for(int j = 0; j < sources.length; j++) {
                      File targetSource = sources[j];
                      String targetSourcePath = targetSource.getAbsolutePath();
                      if(targetSourcePath.equals(historySourcePath)) {
                        foundMatch = true;
                        if(targetSource.lastModified() != sourceHistories[i].getLastModified()) {
                          targetInfo.mustRebuild();
                          break;
                        }
                      }
                    }
                  if(!foundMatch) {
                    targetInfo.mustRebuild();
                  }
              }
            }
          }
        }
    }

    /**
     * This class handles populates the TargetHistory hashtable
     * in response to SAX parse events
     */
    private class TargetHistoryTableHandler extends DefaultHandler {
      private final Hashtable history;
      private String config;
      private String output;
      private long outputLastModified;
      private final Vector sources = new Vector();
      private final File baseDir;

      /**
       * Constructor
       * @param history hashtable of TargetHistory keyed by output name
       * @param outputFiles existing files in output directory
       */
      private TargetHistoryTableHandler(Hashtable history, File baseDir) {
        this.history = history;
        config = null;
        output = null;
        this.baseDir = baseDir;
      }

      /**
       * startElement handler
       */
      public void startElement (String namespaceURI, String localName,
			      String qName, Attributes atts) throws SAXException {
         //
         //   if sourceElement
         //
         if(qName.equals("source")) {
            String sourceFile = atts.getValue("file");
            long sourceLastModified = Long.parseLong(atts.getValue("lastModified"),16);
            sources.addElement(new SourceHistory(sourceFile, sourceLastModified));
        }
        else {
          //
          //   if <target> element,
          //      grab file name and lastModified values
          //      TargetHistory object will be created in endElement
          //
          if(qName.equals("target")) {
            sources.setSize(0);
            output = atts.getValue("file");
            outputLastModified = Long.parseLong(atts.getValue("lastModified"),16);
          }
          else {
            //
            //   if <processor> element,
            //       grab signature attribute
            //
            if(qName.equals("processor")) {
              config = atts.getValue("signature");
            }
          }
        }
      }


      public void endElement (String namespaceURI, String localName,
			    String qName) throws SAXException {
         //
         //   if </target> then
         //       create TargetHistory object and add to hashtable
         //           if corresponding output file exists and
         //           has the same timestamp
         //
         if(qName.equals("target")) {
            if(config != null && output != null) {
              File existingFile = new File(baseDir,output);
              //
              //   if the corresponding files doesn't exist or has a different
              //      modification time, then discard this record
              if(existingFile.exists()) {
                  long existingLastModified = existingFile.lastModified();
                  //
                  //   would have expected exact time stamps
                  //      but have observed slight differences
                  //      in return value for multiple evaluations of
                  //      lastModified().  Check if times are within
                  //      a second
                  long diff = outputLastModified - existingLastModified;
                  if(diff >= -500 && diff <= 500) {
                    SourceHistory[] sourcesArray = new SourceHistory[sources.size()];
                    sources.copyInto(sourcesArray);
                    TargetHistory targetHistory = new TargetHistory(config,output,
                      outputLastModified, sourcesArray);
                    history.put(output,targetHistory);
                  }
              }
            }
            output = null;
            sources.setSize(0);
         }
         else {
           //
           //   reset config so targets not within a processor element
           //      don't pick up a previous processors signature
           //
           if(qName.equals("processor")) {
              config = null;
           }
         }
      }
    }

}
