/*
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2001 The Ant-Contrib project.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "The Jakarta Project", "Ant", and "Apache Software
 *    Foundation" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */
package net.sf.antcontrib.cpptasks.gcc;

import java.io.File;
import java.util.Vector;

import net.sf.antcontrib.cpptasks.CUtil;
import net.sf.antcontrib.cpptasks.compiler.LinkType;
import net.sf.antcontrib.cpptasks.compiler.Linker;
import org.apache.tools.ant.BuildException;

/**
 * Adapter for the GCC C/C++ compiler
 *
 * @author Adam Murdoch
 */
public final class GccCCompiler extends GccCompatibleCCompiler
{
    private String identifier;
    private File[] includePath;
    private boolean isPICMeaningful = true;
    private static final GccCCompiler instance =
      new GccCCompiler("gcc", false,
        new GccCCompiler("gcc", true, null));
    private static final GccCCompiler cppInstance =
      new GccCCompiler("c++",false,
        new GccCCompiler("c++", true, null));
    private static final GccCCompiler gppInstance =
      new GccCCompiler("g++", false,
        new GccCCompiler("g++", true, null));


    /**
     * Private constructor.  Use GccCCompiler.getInstance() to get
     * singleton instance of this class.
     */
    private GccCCompiler(String command,
      boolean isLibtool, GccCCompiler libtoolCompiler) {
      super(command, null, isLibtool, libtoolCompiler);
      isPICMeaningful = System.getProperty("os.name").indexOf("Windows") < 0;
    }

    /**
     * Gets singleton instance of this class
     */
    public static GccCCompiler getInstance() {
      return instance;
    }

    /**
     * Gets singleton instance of this class
     */
    public static GccCCompiler getCppInstance() {
      return cppInstance;
    }

    public void addImpliedArgs(Vector args, boolean debug,
      boolean multithreaded, boolean exceptions, LinkType linkType) {

      super.addImpliedArgs(args,debug,multithreaded,exceptions,
        linkType);
        
      if(isPICMeaningful && linkType.isSharedLibrary()) {
          args.addElement("-fPIC");
      }
    }

    /**
     * Gets singleton instance of this class
     */
    public static GccCCompiler getGppInstance() {
      return gppInstance;
    }

    public File[] getEnvironmentIncludePath() {
      if(includePath == null) {
        //
        //   construct default include path from machine id and version id
        //
        String[] defaultInclude = new String[1];
        StringBuffer buf = new StringBuffer("/lib/");
        buf.append(GccProcessor.getMachine());
        buf.append('/');
        buf.append(GccProcessor.getVersion());
        buf.append("/include");
        defaultInclude[0] = buf.toString();

        //
        //   read specs file and look for -istart and -idirafter
        //
        String[] specs = GccProcessor.getSpecs();
        String[][] optionValues = GccProcessor.parseSpecs(specs,"*cpp:", new String[] { "-isystem ", "-idirafter " });
        //
        //   if no entries were found, then use a default path
        //
        if(optionValues[0].length == 0 && optionValues[1].length == 0) {
          optionValues[0] = new String[] {
            "/usr/local/include",
            "/usr/include",
            "/usr/include/win32api" };
         }
         //
         //  remove mingw entries.
         //    For MinGW compiles this will mean the
         //      location of the sys includes will be
         //      wrong in dependencies.xml
         //      but that should have no significant effect
        for (int i = 0; i < optionValues.length; i++) {
          for (int j = 0; j < optionValues[i].length; j++) {
            if(optionValues[i][j].indexOf("mingw") > 0) {
              optionValues[i][j] = null;
            }
          }
        }
        //
        //   if cygwin then
        //     we have to prepend location of gcc32
        //       and .. to start of absolute filenames to
        //       have something that will exist in the
        //       windows filesystem
        if(GccProcessor.isCygwin()) {
          GccProcessor.convertCygwinFilenames(optionValues[0]);
          GccProcessor.convertCygwinFilenames(optionValues[1]);
          GccProcessor.convertCygwinFilenames(defaultInclude);
        }
        int count = CUtil.checkDirectoryArray(optionValues[0]);
        count += CUtil.checkDirectoryArray(optionValues[1]);
        count += CUtil.checkDirectoryArray(defaultInclude);

        includePath = new File[count];
        int index = 0;
        for (int i = 0; i < optionValues.length; i++) {
          for (int j = 0; j < optionValues[i].length; j++) {
            if (optionValues[i][j] != null) {
              includePath[index++] = new File(optionValues[i][j]);
            }
          }
        }
        for (int i = 0; i < defaultInclude.length; i++) {
          if (defaultInclude[i] != null) {
            includePath[index++] = new File(defaultInclude[i]);
          }
        }
      }
      return (File[]) includePath.clone();
    }


    public String getIdentifier() throws BuildException {
      if (identifier == null) {
        StringBuffer buf;
        if(getLibtool()) {
          buf = new StringBuffer("libtool gcc ");
        } else {
          buf = new StringBuffer("gcc  ");
        }
        buf.append(GccProcessor.getVersion());
        buf.append(' ');
        buf.append(GccProcessor.getMachine());
        identifier = buf.toString();
      }
      return identifier;
    }

    public Linker getLinker(LinkType linkType) {
      return GccLinker.getInstance().getLinker(linkType);
    }
}
