/*
* The Apache Software License, Version 1.1
*
* Copyright (c) 2002 The Ant-Contrib project.  All rights
* reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
*
* 1. Redistributions of source code must retain the above copyright
*    notice, this list of conditions and the following disclaimer.
*
* 2. Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in
*    the documentation and/or other materials provided with the
*    distribution.
*
* 3. The end-user documentation included with the redistribution, if
*    any, must include the following acknowlegement:
*       "This product includes software developed by the
*        Apache Software Foundation (http://www.apache.org/)."
*    Alternately, this acknowlegement may appear in the software itself,
*    if and wherever such third-party acknowlegements normally appear.
*
* 4. The names "The Jakarta Project", "Ant", and "Apache Software
*    Foundation" must not be used to endorse or promote products derived
*    from this software without prior written permission. For written
*    permission, please contact apache@apache.org.
*
* 5. Products derived from this software may not be called "Apache"
*    nor may "Apache" appear in their names without prior written
*    permission of the Apache Group.
*
* THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
* OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
* ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
* USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
* OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
* OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
* SUCH DAMAGE.
* ====================================================================
*
* This software consists of voluntary contributions made by many
* individuals on behalf of the Apache Software Foundation.  For more
* information on the Apache Software Foundation, please see
* <http://www.apache.org/>.
*/
package net.sf.antcontrib.cpptasks.os390;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Vector;

import net.sf.antcontrib.cpptasks.CCTask;
import net.sf.antcontrib.cpptasks.CUtil;
import net.sf.antcontrib.cpptasks.compiler.CommandLineLinker;
import net.sf.antcontrib.cpptasks.compiler.CommandLineLinkerConfiguration;
import net.sf.antcontrib.cpptasks.compiler.LinkType;
import net.sf.antcontrib.cpptasks.compiler.Linker;
import net.sf.antcontrib.cpptasks.types.LibrarySet;
import org.apache.tools.ant.BuildException;

/**
* Adapter for the IBM (R) OS/390 (tm) Linker
*
* @author Hiram Chirino (cojonudo14@hotmail.com)
*/
public final class OS390Linker extends CommandLineLinker
{
    private static final OS390Linker instance =
        new OS390Linker("", "");
    private static final OS390Linker dllLinker =
        new OS390Linker("", ".dll");
    private static final OS390Linker datasetLinker =
        new OS390Linker();

    private String outputPrefix;
    private boolean isADatasetLinker;

    private OS390Linker(String outputPrefix, String outputSuffix)
    {
        super("cxx", "/bogus", new String[] { ".o", ".a", ".lib", ".x" },
          new String[] { ".dll" }, outputSuffix, false, null);
        this.outputPrefix = outputPrefix;
        this.isADatasetLinker = false;
    }

    private OS390Linker()
    {
        super("cxx", "/bogus", new String[] { ".o", ".a", ".lib", ".xds" },
          new String[] { ".dll", ".x" }, ".xds", false, null);
        this.outputPrefix = "";
        this.isADatasetLinker = true;
    }
    

    public static OS390Linker getInstance() {
      return instance;
    }

    public static OS390Linker getDataSetInstance() {
      return datasetLinker;
    }
    

    public File[] getLibraryPath() {
        return CUtil.getPathFromEnvironment("LIB",";");
    }

    public String getOutputFileName(String baseName) {
        return outputPrefix + super.getOutputFileName(baseName);
    }

    protected void addFixed(Boolean fixed, Vector args) {
    }


    protected void addMap(boolean map, Vector args) {
    }

    protected void addBase(long base, Vector args) {
    }

    protected void addIncremental(boolean incremental, Vector args) {
    }


    protected void addStack(int stack, Vector args) {
    }


    public Linker getLinker(LinkType linkType) {
      if ( this	== datasetLinker )
      	return datasetLinker;
      	
      if(linkType.isSharedLibrary()) 
        return dllLinker;
        
      return instance;
    }

    protected void addImpliedArgs(boolean debug, LinkType linkType, Vector args) {
        if(linkType.isSharedLibrary()) {
          args.addElement("-W");
          args.addElement("l,DLL");
        }
    }

    public String getCommandFileSwitch(String commandFile) {
      return "@" + commandFile;
    }

    protected String[] getOutputFileSwitch(CCTask task, String outputFile) {
		if( isADatasetLinker && task.getDataset() !=null ) {
			String ds = task.getDataset();
			outputFile = "//'"+ds+"("+outputFile+")'"; 
		}
		return getOutputFileSwitch(outputFile);
    }

    public String[] getOutputFileSwitch(String outputFile) {
      return new String[] { "-o", outputFile };
    }

    public int getMaximumCommandLength() {
      return Integer.MAX_VALUE;
    }

    public String[] getLibraryPatterns(String[] libnames) {
      return OS390Processor.getLibraryPatterns(libnames);
    }

    public boolean isCaseSensitive() {
      return OS390Processor.isCaseSensitive();
    }

	/*
	 * @see CommandLineLinker#addLibrarySets(LibrarySet[], Vector, Vector, Vector)
	 */
	protected String[] addLibrarySets(CCTask task, LibrarySet[] libsets, Vector preargs, Vector midargs, Vector endargs) {
				
		// If yo want to link against a library sitting in a dataset and
		// not in the HFS, you can just use the //'dataset' notation
		// to specify it. e.g:
		// <libset dir="." libs="//'MQM.V5R2M0.SCSQLOAD'"/> 
		//
		// We have to have special handling here because the file is not 
		// on the normal filesystem so the task will not noramly include it 
		// as part of the link command.
		if( libsets != null ) {
			for( int i=0; i < libsets.length;i++) {
				String libs[] = libsets[i].getLibs();
				for( int j=0; j < libs.length; j++ ) {
					if( libs[j].startsWith("//") ) {
						endargs.addElement("-l");
						endargs.addElement(libs[j]); 
					} else if( libsets[i].getDataset() != null ) {
						String ds = libsets[i].getDataset();
						endargs.addElement( "//'"+ds+"("+libs[j]+")'");
					} 
				}
			}
		}
		
		
		return super.addLibrarySets(task, libsets, preargs, midargs, endargs);
	}
	
	File outputFile;
	CCTask task;
	
	/*
	 * @see CommandLineLinker#link(Task, File, String[], CommandLineLinkerConfiguration)
	 */
	public void link(CCTask task, File outputFile, String[] sourceFiles, CommandLineLinkerConfiguration config)
		throws BuildException {
		this.task = task;
		this.outputFile = outputFile;
		
		if( isADatasetLinker ) {
			int p = outputFile.getName().indexOf(".");
			if( p >= 0 ) {
				String newname = outputFile.getName().substring(0,p);
				outputFile = new File( outputFile.getParent(), newname);
			}
		}		
		super.link(task, outputFile, sourceFiles, config);
	}

	/*
	 * @see CommandLineLinker#runCommand(Task, File, String[])
	 */
	protected int runCommand(CCTask task, File workingDir, String[] cmdline) throws BuildException {
		int rc = super.runCommand(task, workingDir, cmdline);
		// create the .xds file if everything was ok.
		if( rc == 0 ) {
			try {
				outputFile.delete();
				new FileOutputStream( outputFile ).close();
			} catch ( IOException e ) {
				throw new BuildException(e.getMessage());
			}
		}
		return rc;
	}	
}

