/*
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2001 The Ant-Contrib project.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "The Jakarta Project", "Ant", and "Apache Software
 *    Foundation" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */
package net.sf.antcontrib.cpptasks.types;

import java.util.Vector;

import net.sf.antcontrib.cpptasks.CUtil;
import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.types.DataType;
import org.apache.tools.ant.types.Reference;

/**
 * Set of preprocessor macro defines and undefines.
 *
 * @author Mark A Russell <a href="mailto:mark_russell@csgsystems.com">mark_russell@csg_systems.com</a>
 * @author Adam Murdoch
 */
public class DefineSet extends DataType
{
    private Vector defineList = new Vector();

    /** 
     *  A comma-separated list of preprocessor macros to define.
     *  Use nested define elements to define macro values.
     *
     *  @param defList comma-separated list of preprocessor macros
     *  @throws BuildException throw if defineset is a reference  
     */
    public void setDefine(CUtil.StringArrayBuilder defList) throws BuildException
    {
        if ( isReference() ) {
            throw tooManyAttributes();
        }
        addDefines(defList.getValue(), true);
    }

    /** 
     *  A comma-separated list of preprocessor macros to undefine.
     *  
     *  @param defList comma-separated list of preprocessor macros
     *  @throws BuildException throw if defineset is a reference  
     */
    public void setUndefine(CUtil.StringArrayBuilder undefList) throws BuildException
    {
        if ( isReference() ) {
            throw tooManyAttributes();
        }
        addDefines(undefList.getValue(), false);
    }

    /** Adds defines/undefines. */
    private void addDefines(String[] defs, boolean isDefine)
    {
        for ( int i = 0; i < defs.length; i++ ) {
            UndefineArgument def;
            if (isDefine) {
                def = new DefineArgument();
            } else {
                def = new UndefineArgument();
            }
            def.setName(defs[i]);
            defineList.addElement(def);
        }
    }

    /**
     *
     *   Adds a define element.
     *
     *   @throws BuildException if reference
     */
    public void addDefine(DefineArgument arg) throws BuildException
    {
        if ( isReference() ) {
            throw noChildrenAllowed();
        }
        defineList.addElement(arg);
    }

    /**
     *
     *   Adds an undefine element.
     *
     *   @throws BuildException if reference
     */
    public void addUndefine(UndefineArgument arg) throws BuildException
    {
        if ( isReference() ) {
            throw noChildrenAllowed();
        }
        defineList.addElement(arg);
    }

    /**
     *    Specifies that this element should behave as if
     *      the content of the element with the matching id attribute
     *      was inserted at this location.  If specified,
     *      no other attributes or child content should be specified,
     *      other than "description".
     *
     */
    public void setRefid(Reference r) throws BuildException
    {
        if ( !defineList.isEmpty() ) {
            throw tooManyAttributes();
        }
        super.setRefid(r);
    }

    /** Returns the defines and undefines in this set. */
    public UndefineArgument[] getDefines() throws BuildException
    {
        if ( isReference() ) {
            DefineSet defset = (DefineSet)getCheckedRef(DefineSet.class, "DefineSet");
            return defset.getDefines();
        }
        else {
            UndefineArgument[] defs = new UndefineArgument[defineList.size()];
            defineList.copyInto(defs);
            return defs;
        }
    }


    /**
     *    Sets an id that can be used to reference this element.
     *
     *    @param id id
     */
    public void setId(String id) {
        //
        //  this is actually accomplished by a different
        //     mechanism, but we can document it 
        //
    }
    
    /** 
     * Sets a description of the current data type.
     */
    public void setDescription(String desc) {
        super.setDescription(desc);
    }



    public void execute() throws org.apache.tools.ant.BuildException
    {
        throw new org.apache.tools.ant.BuildException("Not an actual task, but looks like one for documentation purposes");
    }

}

