/*
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2001,2002 The Ant-Contrib project.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "The Jakarta Project", "Ant", and "Apache Software
 *    Foundation" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */
package net.sf.antcontrib.cpptasks.types;

import java.io.File;

import net.sf.antcontrib.cpptasks.CUtil;
import net.sf.antcontrib.cpptasks.FileVisitor;
import net.sf.antcontrib.cpptasks.compiler.Linker;
import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.DirectoryScanner;
import org.apache.tools.ant.Project;
import org.apache.tools.ant.ProjectComponent;
import org.apache.tools.ant.types.FileSet;
import org.apache.tools.ant.types.PatternSet;

/**
 * A set of library names.  Libraries can also be added to a link by specifying them in a fileset.
 *
 * For most Unix-like compilers, libset will result in a series of -l and -L 
 *   linker arguments.  For Windows compilers, the library names will be used
 *   to locate the appropriate library files which will be added to the linkers
 *   input file list as if they had been specified in a fileset.
 *
 * 
 *
 * @author Mark A Russell <a href="mailto:mark_russell@csgsystems.com">mark_russell@csg_systems.com</a>
 * @author Adam Murdoch
 * @author Curt Arnold
 */
public class LibrarySet extends ProjectComponent
{
    private final FileSet set = new FileSet();
    private String ifCond;
    private String unlessCond;
    private String[] libnames;
    private boolean explicitCaseSensitive;
    private String dataset;

    public LibrarySet() {
      libnames = new String[0];
    }

    public void setProject(Project project)
    {
        set.setProject(project);
        super.setProject(project);
    }

    /**
     *  Sets the property name for the 'if' condition.
     *
     *  The library set will be ignored unless the property
     *    is defined.
     *
     *  The value of the property is insignificant, but values
     *      that would imply misinterpretation ("false", "no")
     *      will throw an exception when evaluated.
     *
     *  @param propName property name
     */
    public void setIf(String propName)
    {
        ifCond = propName;
    }

    /**
     *  Set the property name for the 'unless' condition. 
     *
     *   If named property is set, the library set will be ignored.
     *
     *   The value of the property is insignificant, but values
     *      that would imply misinterpretation ("false", "no")
     *      of the behavior will throw an exception when evaluated.
     *
     *  @param propName name of property
     */
    public void setUnless(String propName)
    {
        unlessCond = propName;
    }

    protected FileSet getFileSet() {
      return set;
    }

    /**
     *   Library directory.
     *
     *   @param dir library directory
     *
     */
    public void setDir(File dir) throws BuildException {
       set.setDir(dir);
    }

    public File getDir(Project project) {
      return set.getDir(project);
    }

    /**
     * Sets case sensitivity of the file system.  If not set,
     * will default to the linker's case sensitivity.
     *
     * @param isCaseSensitive "true"|"on"|"yes" if file system is case
     *                           sensitive, "false"|"off"|"no" when not.
     */
    public void setCaseSensitive(boolean isCaseSensitive) {
      explicitCaseSensitive = true;
      set.setCaseSensitive(isCaseSensitive);
    }

    /**
     * Comma-separated list of library names without
     * leading prefixes, such as "lib", or extensions,
     * such as ".so" or ".a".
     *
     */
    public void setLibs(CUtil.StringArrayBuilder libs) throws BuildException {
      libnames = libs.getValue();
      for (int i = 0; i < libnames.length; i++) {
        if (libnames[i].indexOf(".") >= 0) {
            throw new BuildException("Library names should not contain periods.");
        }
        if (libnames[i].substring(0,3).equalsIgnoreCase("lib")) {
            throw new BuildException("Library names should not start with \"lib\".");
        }
      }
    }

    public String[] getLibs() {
      String[] retval = (String[]) libnames.clone();
      return retval;
    }

    /**
     * Returns true if the define's if and unless conditions
     * (if any) are satisfied.
     */
    public boolean isActive(org.apache.tools.ant.Project p) {
        if(p == null) {
           throw new NullPointerException("p");
        }
        if (ifCond != null) {
          String ifValue = p.getProperty(ifCond);
          if(ifValue != null) {
            if(ifValue.equals("no") || ifValue.equals("false")) {
              throw new BuildException("property " + ifCond +
                " used as if condition has value " + ifValue +
                " which suggests a misunderstanding of if attributes");
            }
          } else {
          	return false;
          }
        }
        if (unlessCond != null) {
          String unlessValue = p.getProperty(unlessCond);
          if(unlessValue != null) {
            if(unlessValue.equals("no") || unlessValue.equals("false")) {
              throw new BuildException("property " + unlessCond +
                " used as unless condition has value " + unlessValue +
                " which suggests a misunderstanding of unless attributes");
            }
            return false;
          }
        }
        return true;
    }

    public void visitLibraries(Project project,
      Linker linker,
      File[] libpath,
      FileVisitor visitor) throws BuildException {


      FileSet localSet = (FileSet) set.clone();

      //
      //   unless explicitly set
      //      will default to the linker case sensitivity
      //
      if(!explicitCaseSensitive) {
        boolean linkerCaseSensitive = linker.isCaseSensitive();
        localSet.setCaseSensitive(linkerCaseSensitive);
      }

      //
      //  if there was a libs attribute then
      //     add the corresponding patterns to the FileSet
      //
      if(libnames != null && libnames.length > 0) {
        String[] patterns = linker.getLibraryPatterns(libnames);
        for (int i = 0; i < patterns.length; i++) {
          PatternSet.NameEntry entry = localSet.createInclude();
          entry.setName(patterns[i]);
        }
      }

      //
      //  if there was no specified directory then
      //     run through the libpath backwards
      //
      if(localSet.getDir(project) == null) {
        //
        //  scan libpath in reverse order
        //     to give earlier entries priority
        //
        for(int j = libpath.length-1; j >= 0; j--) {
          FileSet clone = (FileSet) localSet.clone();
          clone.setDir(libpath[j]);
          DirectoryScanner scanner = clone.getDirectoryScanner(project);
          File basedir = scanner.getBasedir();
          String[] files = scanner.getIncludedFiles();
          for ( int k = 0; k < files.length; k++ ) {
            visitor.visit(basedir, files[k]);
          }
        }
      } else {
          DirectoryScanner scanner = localSet.getDirectoryScanner(project);
          File basedir = scanner.getBasedir();
          String[] files = scanner.getIncludedFiles();
          for ( int k = 0; k < files.length; k++ ) {
            visitor.visit(basedir, files[k]);
          }
      }
    }

	/**
	 * Gets the dataset.  Used on OS390 if the libs are in a dataset.
	 * @return Returns a String
	 */
	public String getDataset() {
		return dataset;
	}

	/**
	 * Sets the dataset.  Used on OS390 if the libs are in a dataset.
	 * @param dataset The dataset to set
	 */
	public void setDataset(String dataset) {
		this.dataset = dataset;
	}

    public void execute() throws org.apache.tools.ant.BuildException
    {
        throw new org.apache.tools.ant.BuildException("Not an actual task, but looks like one for documentation purposes");
    }

}

